#include <conio.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <mem.h>
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include "lgc_head.h"
#include "lgc_def.h"
#include "bafulo.h"
#include "help.h"

char screen_buffer[ SAVE_SCREENS ][ 80*25*2 ];
char currently_saved_screens = 0;
unsigned char beep_or_not = '\0';

extern char lgc_name[ LGCNAMELEN + 1 ];

/*
   Name       : read_in.
   Function   : get a string from keyboard until <RETURN> is received of
		at most 'len'-1 characters. Special characters are ignored.
		Characters are displayed on the screen.
*/
char *read_in( int len, char *init_string, int *exit_character )
{
   static char buffer[ 80 ];  /* That should do for the whole program. */
   int c;
   int i = 0;

   if ( len > 80 )
   {
      printf( "BAFULO.C, function read_in: Required buffer %d", len );
      exit( 1 );
   }
   len = ( len > 1000 ? 999 : len );
   if ( init_string != NULL && ( i = strlen( init_string ) ) > 0 )
   {
      if ( i > len )
	 init_string[ len ] = '\0';   /* We brutally cut it off. */
      cprintf( "%s", init_string );
      strcpy( buffer, init_string );
   }
   do
   {
      c = get_key();
      if ( c == BACKSPACE  && i > 0 )
      {
	 putch( c ); putch( ' ' ); putch( c ); /* Erase last character. */
	 i -= 1;
      }
      else if ( isprint( c ) && i < ( len - 1 ) )
      {
	 buffer[ i ] = c;
	 putch( c );
	 i += 1;
      }
      else if ( c == ESC || c == TAB || c == SHIFT_TAB || c == UP || c == DOWN ||
      c == LEFT || c == RIGHT )
      {
	 buffer[ i ] = '\0';
	 *exit_character = c;
	 return ( buffer );
      }
      else if ( c != RETURN && c != 0 ) beep();
   }
   while ( c != RETURN );
   buffer[ i ] = '\0';
   *exit_character = RETURN;
   return buffer;
} /* read_in */

/*
   Name       : re1.
   Function   : checks if 'str' matches the regexpr [a-z][a-z0-9_]*,
		returns the "bad" position, -2 on success,
		-1 if 'str' is empty.
*/
int re1( char *str )
{
   int i;
   char j;

   if ( strlen( str ) == 0 ) return ( -1 );
   if ( !isalpha( *str ) || !islower( *str ) )
      return ( 0 );
   for ( i = 1; i < strlen( str ); i++ )
   {
      j = *( str + i );
      if ( !( isdigit( j ) || ( isalpha( j ) && islower( j ) ) || j == '_' ) )
	 return ( i );
   }
   return ( -2 );
} /* re1 */

/*
   Name       : re2.
   Function   : checks if 'str' matches the regexpr
		([a-z][a-z0-9_]*|[-+*^<>=~?@#$&]+|0|[1-9][0-9]*),
		returns the "bad" position, -2 on success,
		-1 if 'str' is empty.
*/
int re2( char *str )
{
   int i;
   char j;

   if ( strlen( str ) == 0 ) return ( -1 );
   if ( isalpha( *str ) && islower( *str ) )
   {
      for ( i = 1; i < strlen( str ); i++ )
	 if ( !isalnum( *( str + i ) ) && *( str + i ) != '_' )
	    return ( i );
      return ( -2 );
   }
   else
   {
      switch( *str )
      {
	 case '0':
	    if ( strlen( str ) == 1 ) return ( -2 );
	    else return ( 0 );
	 case '1':
	 case '2':
	 case '3':
	 case '4':
	 case '5':
	 case '6':
	 case '7':
	 case '8':
	 case '9':
	    for ( i = 1; i < strlen( str ); i++ )
	       if ( !isdigit( *( str + i ) ) ) return ( i );
	    return ( -2 );
	 case '-':
	 case '+':
	 case '*':
	 case '^':
	 case '<':
	 case '>':
	 case '=':
	 case '~':
	 case '?':
	 case '@':
	 case '#':
	 case '$':
	 case '&':
	    for ( i = 1; i < strlen( str ); i++ )
	    {
		j = *( str + i );
		if ( j != '-' && j != '+' && j != '*' && j != '^'
		&& j != '<' && j != '>' && j != '=' && j != '~'
		&& j != '?' && j != '@' && j != '#' && j != '$'
		&& j != '&' ) return ( i );
	    }
	    return -2;
	 default: return ( 0 );
      }
   }
} /* re2 */

/*
   Name       : re3.
   Function   : checks if 'str' matches the regexpr [0-9]+, returns the
		"bad" position, -1 if 'str' is empty, -2 on success.
*/
int re3( char *str )
{
   int i;
   char j;

   if ( strlen( str ) == 0 ) return ( -1 );
   for ( i = 0; i < strlen( str ); i++ )
   {
      j = *( str + i );
      if ( !( isdigit( j ) || ( i == 0 && j == '-' ) ) ) return ( i );
   }
   return ( -2 );
} /* re3 */

/*
   Name       : display_error.
   Function   : Displays 'msg' on the screen.
*/

void display_error( char *msg )
{
   int i, j;

   i = ( strlen( msg ) >= 40 ? 40 : strlen( msg ) ) / 2 + 1;
   j = strlen( msg )/40 + 1;
   save_screen();
   ( void ) create_window( 40 - i, 12, 41 + i, 13 + j, RED, WHITE,
			   DOUBLE_LINE_BOX, " Error " );
   gotoxy( 3, 2 );
   display_auto_wrap( msg, 40 );
   cursor( 0 );
   wait_for( ESC );
   cursor( 1 );
   restore_screen();
} /* display_error */

/*
   Name       : create_window.
   Function   : Put a window on the screen and return the upper left
		and the lower right corner as window_borders.
*/

struct window_borders create_window( int x1, int y1, int x2, int y2, int bkc,
				     int fc, int boxtype, char *title )
{
   struct window_borders dwb;

   window( x1, y1, x2, y2 );
   TATTR( bkc, fc );
   clrscr();
   if ( boxtype != NO_BOX )
   {
      window( x1, y1, x2, y2 + 1 );
      box( 1, 1, 1 + x2 - x1, 1 + y2 - y1, boxtype );
   }
   window( x1, y1, x2, y2 );
   if ( title != NULL && strlen( title ) )
      OUTXY( max( ( x2 - x1 + 1 - strlen( title ) )/2 + 1, 1 ), 1, title );
   dwb.x1 = x1;
   dwb.x2 = x2;
   dwb.y1 = y1;
   dwb.y2 = y2;
   dwb.bkg = bkc;
   dwb.fg = fc;
   return( dwb );
}

/*
   Name       : display_about.
   Function   : about the Logic Creator.
*/

void display_about()
{
   welcome_window();
   main_window();
} /* display_about */

/*
   Name       : wait_for.
   Function   : wait until the given character "which" is pressed.
*/

void wait_for( char which )
{
   int i;
   do
   {
      i = get_key();
   }
   while ( i != which );
}

/*
   Name       : esc_or_return.
   Function   : return as soon as one of the given characters is typed.
		Return the character typed.
*/
void esc_or_return()
{
   int n;

   do {
      cursor( 0 );
      n = get_key();
      cursor( 1 );
   }
   while ( n != RETURN && n != ESC );
} /* esc_or_return */

/*
   Name       : save_screen.
   Function   : saves the current state of the total screen.
*/

void save_screen()
{
   if ( currently_saved_screens == SAVE_SCREENS )
   {
      printf( "BAFULO.C, function save_screen: Increase defined SAVE_SCREENS." );
      exit( 0 );
   }
   movedata( CLR_BASE, 0, _DS,
	     ( unsigned ) screen_buffer[ currently_saved_screens ],
	     80*25*2 );
   currently_saved_screens += 1;
} /* save_screen */

/*
   Name       : restore_screen.
   Function   : restores the saved screen.
*/

void restore_screen()
{
   currently_saved_screens -= 1;
   movedata( _DS, ( unsigned ) screen_buffer[ currently_saved_screens ],
	     CLR_BASE, 0, 80*25*2 );
} /* restore_screen */

/*
   Name       : inverse.
   Function   : Writes "string" with background & foreground inverse.
*/

void inverse( int wherex, int wherey, char *str1, int modus )
{
   gotoxy( wherex, wherey );
   textattr( modus );
   cprintf( "%s", str1 );
} /* inverse */

/*
   Name       : in_red.
   Function   : Writes "string" in red at the current position and changes
		back to "old_color" afterwards.
*/

void in_red( char *str1, int old_color )
{
   textcolor( RED );
   cprintf( "%s", str1 );
   textcolor( old_color );
} /* in_red */

/*
   Name       : screen.
   Function   : focus on the total screen.
*/

void screen()
{
   window( 1, 1, 80, 25 );
} /* screen */

/*
   Name       : main_window.
   Function   : returns to the main window and sets the main window colors.
*/

void main_window()
{
   window( 2, 3, 79, 2 + MAIN_WINDOW_LENGTH );
   gotoxy( 1, 1 );
   TATTR( BLACK, LIGHTGRAY );
   clrscr();
} /* main_window */

/*
   Name       : welcome_window.
   Function   : display a welcome at the beginning and wait for a keystroke.
*/

void welcome_window()
{
   (void) create_window( 12, 5, 68, 19, LIGHTGRAY, BLACK, DOUBLE_LINE_BOX, 0 );
   (void) create_window( 13, 6, 67, 18, LIGHTGRAY, BLACK, NO_BOX, 0 );
   TATTR( LIGHTGRAY, BLUE ); cprintf( "%s",
/*
"                   ########|       ######|   ######|  \n\r"
"                     ####\'/         `####|   ####\'/ \n\r"
"                     ####|           ####|  ,####|    \n\r"
"    ,##########.\\    ####|           ####| ####\'/   \n\r"
"   ####(     ####|   ####|           ########\'/      \n\r"
"   ##############|   ####|           ########.\\      \n\r"
"   ####\'/            ####|           ####|`####.\\   \n\r"
"   ####|             ####|    ,##|   ####|  `####|    \n\r"
"   ####.\\    ####|   ####.\\ ,####|  ,####|   ####.\\\n\r"
"    `##########\'/  ##############| ######|   ######| ", BLACK );
*/
"                                  \n\r"
"         .                                \n\r"
"                        |                 \n\r"
"     ۲۲۲      ۲ --O--      ۲ ۲        \n\r"
"   ۲      ۲    ۲   |        ۲۲۲         \n\r"
"               .      *     \n\r"
"                   *                  \n\r"
"              .                     \n\r"
"                      .       \n\r"
"                  " );
   TATTR( LIGHTGRAY, BLACK );
   OUTXY( 3, 12,
      "*L*ogic *K*reator (c) by \\/\\//\\/   -   Version 0.9" );
/*
   OUTREDXY( 9, 3, "         L O G I C - C R E A T O R", BLACK );
   OUTXY( 9, 5,    "           (c) 1995 by \\/\\//\\/\n\r"
		   "                Version 1.0\n\n\r"
		   "          Hit any key to proceed." );
*/
   cursor( 0 );
   do { } while ( !kbhit() );
   cursor( 1 );
}

/*
  Name        : get_key.
  Function    : returns the value of a key pressed on the keyboard. If it
		is a non ASCII character, 256 is added.
*/

int get_key()
{
   int k;

   k = getch();

   if ( k == 0 )
   {
       k = getch();
       k += 256;
   } /* if */
/*   printf( " K: %d ", k ); */
   if ( k == F1 )
   {
      display_help();
      return( 0 );
   }
   return ( k );

} /* get_key */

/*
   Name       : box
   Function   : makes a box around specified area.
*/

void box( int xx1, int yy1, int xx2, int yy2, int line_type )
{
   int i;
   char lines[ 3 ][ 6 ][ 2 ] = { "","","","","","",
				 "","","","","","",
				 "/","\\","\\","/","-","|" };

   if ( line_type != SINGLE_LINE_BOX && line_type != DOUBLE_LINE_BOX )
      return;
   gotoxy( xx1, yy1 );
   cprintf( "%s", lines[ line_type ][ 0 ] );
   for ( i = 1; i < ( xx2 - xx1 ); i++ )
      cprintf( "%s", lines[ line_type ][ 4 ] );
   cprintf( "%s", lines[ line_type ][ 1 ] );
   for ( i = 1; i < ( yy2 - yy1 ); i++ )
   {
      gotoxy( xx1, yy1 + i );
      cprintf( "%s", lines[ line_type ][ 5 ] );
      gotoxy( xx2, yy1 + i );
      cprintf( "%s", lines[ line_type ][ 5 ] );
   }
   gotoxy( xx1, yy2 );
   cprintf( "%s", lines[ line_type ][ 2 ] );
   for ( i = 1; i < ( xx2 - xx1 ); i++ )
      cprintf( "%s", lines[ line_type ][ 4 ] );
   cprintf( "%s", lines[ line_type ][ 3 ] );
   gotoxy( xx1, yy1 );
} /* box */

/*
   Name       : strpad;
   Function   : pads str with spaces to the right until strlen( str ) ==
		length.
*/
char *strpad( char *str, int length )
{
   static char dummy[ 80 ]; /* That should do ... */
   char format[ 20 ];

   if ( length >= 100 )
   {
      printf( "BAFULO.C, function strpad: Need longer dummy for string %s", str );
      exit( 1 );
   }
   sprintf( format, "%%-%ds", ( length < 0 ? 0 : length ) );
   sprintf( dummy, format, str );
   return dummy;
} /* strpad */

/*
   Name       : draw_slide_bar.
   Function   : used by scroll_menu. Main purpose: to signify with up/down
		arrows that there is more.
*/

void draw_slide_bar( int how_many, int window_length, int pos,
		     int scroll_pos )
{
   int i;

   for ( i = 1; i <= window_length; i++ )
   {
      gotoxy( 1, i );
      cprintf( "%c", 177 );
   }
   if ( pos - scroll_pos >= 0 )
   {
      gotoxy( 1, 1 );
      cprintf( "%c", 30 );
   }
   if ( pos + window_length - scroll_pos < how_many - 1 )
   {
      gotoxy( 1, window_length );
      cprintf( "%c", 31 );
   }
}

/*
   Name       : scroll_menu.
   Function   : scrolls the "**str" in the given window until one is selected
		with RETURN or aborted with ESC.
*/
int scroll_menu( char **str, int how_many, int wx, int wy, int window_length,
		 int pos_init, int bkg, int fg, int boxtype )
{
   int c;
   int i, pos, scroll_pos;
   char dummy[ 10 ];
   int width = 0;
   int z;

   for ( i = 0; i < how_many; i++ )
      if ( strlen( str[ i ] ) > width ) width = strlen( str[ i ] );
   if ( how_many < window_length ) window_length = how_many;
   z = ( how_many == window_length ? 1 : 2 + !!( boxtype != NO_BOX ) );
   if ( boxtype != NO_BOX )
   {
      ( void ) create_window( wx, wy, wx + width + z + 1,
			      wy + window_length + 1, bkg, fg, boxtype, 0 );
      wx += 1;
      wy += 1;
      if ( z != 1 )
      {
	 gotoxy( z, 1 );
	 cprintf( "%c", ( boxtype == SINGLE_LINE_BOX ? '' : '' ) );
	 gotoxy( z, window_length + 2 );
	 cprintf( "%c", ( boxtype == SINGLE_LINE_BOX ? '' : '' ) );
      }
   }
   ( void ) create_window( wx, wy, wx + width + z - 1, wy - 1 + window_length,
		  bkg, fg, NO_BOX, 0 );
   if ( z != 1 )
      for ( i = 1; i <= window_length; i++ )
      {
	 gotoxy( 2, i );
	 cprintf( "" );
      }
   if ( ( pos = pos_init ) >= how_many ) pos = how_many - 1;
   if ( pos <= window_length ) scroll_pos = pos + 1;
   else scroll_pos = window_length;
   if ( pos < window_length ) c = pos;
   else c = window_length;
   for ( i = 0; i < window_length && ( i + pos - c ) < how_many; i++ )
   {
      gotoxy( z, 1 + i );
      cprintf( "%s", str[ i + pos - c ] );
   }
   if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
   strcpy( dummy, ( z != 1 ? "\b%s" : "%s" ) );
   do
   {
      inverse( z, scroll_pos, strpad( str[ pos ], width ), fg*16 + bkg );
      cursor( 0 );
      c = get_key();
      cursor( 1 );
      inverse( z, scroll_pos, strpad( str[ pos ], width ), fg + bkg*16 );
      switch( c ) {
      case UP:
	 if ( scroll_pos == 1 )
	 {
	    if ( pos > 0 )
	    {
	       pos -= 1;
	       gotoxy( z, 1 );
	       insline();
	       cprintf( dummy, strpad( str[ pos ], width ) );
	    }
	 }
	 else
	 {
	    pos -= 1;
	    scroll_pos -= 1;
	 }
         if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
	 break;
      case DOWN:
	 if ( scroll_pos >= window_length )
	 {
	    if ( pos < ( how_many - 1 ) )
	    {
	       pos += 1;
	       cprintf( "\n" );
	       gotoxy( z, window_length );
	       cprintf( dummy, strpad( str[ pos ], width ) );
	    }
	 }
	 else if ( pos < ( how_many - 1 ) )
	 {
	    pos += 1;
	    scroll_pos += 1;
	 }
         if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
	 break;
      case PGUP :
	 if ( scroll_pos > 1 )
	 {
	    pos -= ( scroll_pos - 1 );
	    scroll_pos = 1;
	 }
	 else
	 {
	    if ( pos != 0 )
	    {
	       if ( ( pos -= window_length ) < 0 ) pos = 0;
	       gotoxy( z, 1 );
	       textattr( bkg*16 + fg );
	       for ( i = 0; i < window_length && ( i + pos ) < how_many; i++ )
	       {
		  gotoxy( z, 1 + i );
		  cprintf( dummy, strpad( str[ pos + i ], width ) );
	       }
	    }
	 }
         if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
	 break;
      case PGDOWN :
	 if ( scroll_pos < window_length && scroll_pos < how_many )
	 {
	    if ( window_length > how_many )
	    {
	       scroll_pos = how_many;
	       pos = how_many - 1;
	    }
	    else
	    {
	       pos += ( window_length - scroll_pos );
	       scroll_pos = window_length;
	    }
	 }
	 else
	 {
	    if ( pos != how_many - 1 )
	    {
	       if ( ( pos += window_length ) >= how_many )
		  pos = how_many - 1;
	       gotoxy( z, 1 );
	       textattr( bkg*16 + fg );
	       for ( i = 0; i < window_length; i++ )
	       {
                  gotoxy( z, 1 + i );
		  cprintf( dummy, strpad( str[ pos - scroll_pos + 1 + i ],
			   width ) );
	       }
	    }
	 }
         if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
	 break;
      case POS1:
	 scroll_pos = 1;
	 if ( pos != 0 )
	 {
	    pos = 0;
	    gotoxy( z, 1 );
	    textattr( bkg*16 + fg );
	    for ( i = 0; i < window_length && ( i + pos ) < how_many; i++ )
	    {
               gotoxy( z, 1 + i );
	       cprintf( dummy, strpad( str[ pos + i ], width ) );
	    }
	 }
         if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
	 break;
      case END:
	 if ( ( scroll_pos = window_length ) > how_many )
	    scroll_pos = how_many;
	 if ( pos != how_many - 1 )
	 {
	    pos = how_many - 1;
	    gotoxy( z, 1 );
	    textattr( bkg*16 + fg );
	    for ( i = 0; i < window_length; i++ )
	    {
	       gotoxy( z, 1 + i );
	       cprintf( dummy, strpad( str[ pos - scroll_pos + 1 + i ],
			width ) );
	    }
	 }
         if ( z != 1 ) draw_slide_bar( how_many, window_length, pos, scroll_pos );
	 break;
      case RETURN:
	 return( pos );
      }
   }
   while ( c != ESC );
   return( -1 );
} /* scroll_menu */

/*
   Name       : read_in_string.
   Function   : reads in a string of at most strlen "length" and checks if
		it matches regular expression "re". If re == 0 or higher
		than 3 no match is tested.
*/

char *read_in_string( int wx, int wy, int length, int re, int bkg, int fg,
		      char *init_string, int *exit_character )
{
   int num;
   static char buffer[ 80 ];
   if ( length > 79 )
   {
      printf( "BAFULO.C, function read_in_string: Need buffer string of length %d.", length );
      exit( 1 );
   }
   do
   {
      (void) create_window( wx, wy, wx + length, wy, bkg, fg, NO_BOX, 0 );
      cprintf( ">" );
      strcpy( buffer, read_in( length, init_string, exit_character ) );
      num = -2;
      switch( re )
      {
	 case 0: break;
	 case 1: num = re1( buffer ); break;
	 case 2: num = re2( buffer ); break;
	 case 3: num = re3( buffer ); break;
      }
      if ( num == -1 )
      {
	 clrscr();
	 cprintf( ">" );
	 display_error( "Enter a non-empty string." );
      }
      else if ( num >= 0 )
      {
	 sprintf( buffer, "Error at position %d.", num + 1 );
	 display_error( buffer );
      }
   }
   while ( num != -2 );
   return buffer;
} /* read_in_string */

/*
   Name       : read_in_int.
   Function   : read in an integer in a speficied window, do it until
		it is an int between "min" and "max" inclusively. Returns
		the value in variable 'ret'.
*/

void read_in_int( int *ret, int wx, int wy, int digits, int mini, int maxi,
		  int bkg, int fg, int init_int, int *exit_character )
{
   int c, cc;
   struct window_borders wb;
   char format[ 20 ];

   wb = create_window( wx, wy, wx + 1 + digits, wy, bkg, fg, NO_BOX, 0 );
   wb.x2 += 1;
   focus_on_window( wb );
   *ret = ( init_int < mini ? mini : init_int );
   sprintf( format, "%%s%%%dd%%s", digits );
   cprintf( format, ( *ret > mini ? "\x11" : " " ), *ret,
	       ( *ret < maxi ? "\x10" : " " ) );
   gotoxy( 1 + digits, 1 );
   do
   {
      c = get_key();
      cc = c - (int) '0';
      if ( isdigit( c ) && cc >= mini && cc <= maxi )
	 *ret = cc;
      else
	 switch( c )
	 {
	    case LEFT: if ( *ret > mini ) *ret -= 1; break;
	    case RIGHT: if ( *ret < maxi ) *ret +=1; break;
	 }
      gotoxy( 1, 1 );
      cprintf( format, ( *ret > mini ? "\x11" : " " ), *ret,
	       ( *ret < maxi ? "\x10" : " " ) );
      gotoxy( 1 + digits, 1 );
   }
   while ( c != RETURN && c != ESC && c != TAB && c != SHIFT_TAB &&
   c != UP && c != DOWN );
   *exit_character = c;
} /* read_in_int */

/*
   Name       : focus_on_window.
   Function   : Focus on a given window defined by "wb".
*/

void focus_on_window( struct window_borders wb )
{
   window( wb.x1, wb.y1, wb.x2, wb.y2 );
   TATTR( wb.bkg, wb.fg );
} /* focus_on_window */

/*
   Name       : horizontal_select_menu.
   Function   : select one of the given string-options and return the
		index of the chosen one.
*/

int horizontal_select_menu( int wx, int wy, int length, int just_init,
			    int out, int init_pos, int how_many, ... )
{
   va_list vl;
   int i, width = 0, wwidth, space, pos;
   char str[ 10 ][ 25 ];
   char format[ 80 ], dummy[ 80 ];

   if ( how_many > 10 )
   {
      printf( "Horizontal_select_menu: Need more space for entries." );
      exit( 1 );
   }
   va_start( vl, how_many );
   for ( i = 0; i < how_many; i++ )
   {
      strcpy( str[ i ], va_arg( vl, char * ) );
      if ( strlen( str[ i ] ) > width ) width = strlen( str[ i ] );
   }
   for ( i = 0; i < how_many; i++ )
   {
      sprintf( format, "%%%ds",
	       ( width + strlen( str[ i ] ) )/2 );
      strcpy( dummy, str[ i ] );
      sprintf( str[ i ], format, dummy );
   }
   sprintf( format, "<%%-%ds>", width );
   for ( i = 0; i < how_many; i++ )
   {
      strcpy( dummy, str[ i ] );
      sprintf( str[ i ], format, dummy );
   }
   va_end( vl );
   wwidth = ( width += 2 ) * how_many;
   (void) create_window( wx, wy, wx + max( length, wwidth ) - 1, wy,
			 LIGHTGRAY, BLACK, NO_BOX, 0 );
   space = max( length - wwidth, 0 )/( how_many + 1 );
   for ( i = 0; i < how_many; i++ )
   {
      gotoxy( wherex() + space, 1 );
      cprintf( str[ i ] );
   }
   if ( just_init ) return( -ESC );
   if ( init_pos < 0 ) pos = 0;
   else if ( init_pos >= how_many ) pos = how_many - 1;
   else pos = init_pos;
   do
   {
      inverse( 1 + space + ( space + width )*pos, 1, str[ pos ], LIGHTGRAY );
      cursor( 0 );
      i = get_key();
      cursor( 1 );
      inverse( 1 + space + ( space + width )*pos, 1, str[ pos ], LIGHTGRAY*16 );
      switch( i )
      {
	 case RETURN:
	    return( pos );
	 case DOWN:
	 case UP:
	    if ( out ) return( -i );
	 case TAB:
	 case RIGHT:
	    if ( ( pos += 1 ) >= how_many )
	    {
	       if ( out ) return( -i );
	       else pos = 0;
	    }
	    break;
	 case SHIFT_TAB:
	 case LEFT:
	    if ( ( pos -= 1 ) < 0 )
	    {
	       if ( out ) return( -i );
	       else pos = how_many - 1;
	    }
	    break;
	 case ESC: return( -ESC );
      }
   }
   while ( 1 );
} /* horizontal_select_menu */

/*
   Name       : ok_or_cancel.
   Function   : paste a ok/cancel menu on the screen.
*/

int ok_or_cancel( int wx, int wy, int length, int just_init, int out,
		  int init_pos )

{
   return( horizontal_select_menu( wx, wy, length, just_init, out, init_pos, 3,
				   "Accept", "Cancel", "Edit" ) );
}

/*
   Not in use yet... rather never.
*/

void cursor( int modus )
{
   union REGS rg;

   rg.h.ah = 2;
   rg.h.bh = 0;
   rg.h.dh = ( modus ? 0 : 25 );
   rg.h.dl = ( modus ? 0 : 80 );

   int86( 0X10, &rg, &rg );
}

void beep()
{
   if ( beep_or_not ) putch( '\a' );
}

int toggle_beep()
{
   return( beep_or_not = (char) !beep_or_not );
}

int confirm( char *message, int how_many )
{
   int i, maxi, hy;
   struct window_borders wb;

   save_screen();
   hy = strlen( message )/40 + 1;
   maxi = 6 + max( strlen( message ) < 40 ? strlen( message ) : 40,
	  9*how_many );
   wb = create_window( 39 - maxi/2, 10, 41 + maxi/2, 13 + hy, LIGHTGRAY,
	WHITE, DOUBLE_LINE_BOX, " Confirm " );
   for ( i = 0; i <= hy; i++ )
      OUTREDXY( 3, i + 2, "۳", BLACK );
   OUTREDXY( 3, i + 2, ")", BLACK );
   gotoxy( 7, 2 );
   TATTR( LIGHTGRAY, BLACK );
   display_auto_wrap( message, 40 );
   i = horizontal_select_menu( wb.x1 + 5, wb.y2 - 1, wb.x2 - wb.x1 - 7,
       !JUST_INIT, !OUT, NO_INIT_POS, how_many, "  Ok  ", "Cancel" );
   restore_screen();
   return( i );
} /* confirm */

void display_auto_wrap( char *str, int ww )
{
   char *s, *t, c;
   int startx;

   s = str;
   startx = wherex();
   while ( *s != '\0' )
   {
      t = s;
      for ( ; *t != '\0' && *t != ' '; t++ );
      if ( *t == ' ' ) t += 1;
      c = *t;
      *t = '\0';
      if ( wherex() + strlen( s ) - startx >= ww )
	 gotoxy( startx, wherey() + 1 );
      cprintf( "%s", s );
      *t = c;
      s = t;
   }
} /* display_auto_wrap */
